/*   viewer3d.c
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*            National Center for Biotechnology Information (NCBI)
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government do not place any restriction on its use or reproduction.
*  We would, however, appreciate having the NCBI and the author cited in
*  any work or product based on this material
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
* ===========================================================================
*
* File Name:  viewer3d.c
*
* Author:  Alex Smirnov
*
* Version Creation Date:   04/03/95
*
* $Revision: 5.10 $
*
* File Description: 
*
* Modifications:  
* --------------------------------------------------------------------------
* Date     Name        Description of modification
* -------  ----------  -----------------------------------------------------
* $Log: viewer3d.c,v $
 * Revision 5.10  1996/11/22  19:47:29  vakatov
 * Nlm_CreateViewer3D():  set image-dependent background color;
 * + extensive type casting
 *
 * Revision 5.9  1996/09/20  13:54:47  vakatov
 * Nlm_AttachPicture3D(): assign new picture *at first*; then do anyth. else
 *
 * Revision 5.8  1996/09/19  18:54:18  vakatov
 * Replace Reset() by CorrectBarValue() to avoid undesirable call to
 * the scrollbar callbacks
 *
 * Revision 5.7  1996/07/25  14:52:34  epstein
 * BigScalar change for OSF1 portability
 *
 * Revision 5.6  1996/07/24  16:59:59  vakatov
 * Trapped crashes caused by attempts to rotate and zoom blank 3D-pictures
 *
 * Revision 5.5  1996/07/23  21:51:29  vakatov
 * [WIN_MOTIF]  Reserve 32th(instead of 0th) color for the background
 * color; it lets CN3D do not confuse other parts of the program and allows
 * to save the background color to GIF metafile
 *
 * Revision 5.4  1996/06/19  18:39:43  kans
 * *** empty log message ***
 *
 * Revision 5.3  1996/06/14  20:16:38  vakatov
 * [WIN_MSWIN]  Slight modification in the bg-color setting
 *
 * Revision 5.2  1996/06/14  14:33:04  vakatov
 * Added a set of functions to control the 3D-Viewer background color
 * [WIN_MOTIF] Special precautions made to avoid Motif/3D-Viewer color conflict
 *
 * Revision 5.1  1996/06/07  15:19:39  hogue
 * Removed the PlayLayer button and sundry stuff from MAC.
 *
 * Revision 5.0  1996/05/28  13:45:08  ostell
 * Set to revision 5.0
 *
 * Revision 1.23  1996/05/23  14:36:35  hogue
 * Fixed PrintViewer/PasteViewer discrepancy in .h file, added
 * IsPlaying3D, StopPlaying3D, StartPlaying3D for more
 * external control of the animation
 *
 * Revision 1.22  1996/05/21  22:00:40  vakatov
 * The viewer scrollbar width/height are now treated more carefully.
 * Do not store viewer width and height in the camera position storage as
 * it caused a coredump when resizing the viewer.
 * Viewer rotation flag is ON(was OFF) in the very begining.
 *
 * Revision 1.21  1996/05/14  15:29:08  hogue
 * Made the Horizontal viewer controls more compact.
 *
 * Revision 1.20  1996/05/13  15:04:21  hogue
 * Save button in Viewer Controls now optional
 *
 * Revision 1.19  1996/05/09  18:45:35  vakatov
 * There was a bug resulting the program crash while switching between
 * pictures with different number of layers(and then restoring the
 * camera position) -- fixed.
 *
 * Revision 1.18  1996/05/07  18:30:03  vakatov
 * Nlm_CreateViewer3D() -- slightly changed;
 * Nlm_AllLayerSetStates3D() function -- added;
 * Nlm_AllLayerSet3D() -- modified, "state" parameter added
 * Nlm_RewindLayer3D() -- prototype and content modified;
 * + a lot of variable casting and code simplifying
 *
 * Revision 1.17  1996/04/23  16:30:47  vakatov
 * Nlm_PlayLayer3D() -- fixed an extra(some cycles) delay responding to
 * the V3D_bPLAY("Go") toggle button release event
 *
 * Revision 1.16  1996/04/17  21:31:58  vakatov
 * Implemented "Nlm_SaveViewer3D()" function which allows one to save
 * presently visible part of the 3D-viewer picture into a GIF metafile.
 * Added "Copy" button to the 3D-viewer controls.
 * Fixed a bug caused by "the asynchroneous commenting" of "Print" and "Copy"
 * controls.
 * Added "Nlm_GetViewerImage3D()" function returning image of the 3D-viewer.
 *
 * Revision 1.15  1996/04/09  15:24:49  vakatov
 * Most of the basic, commonly used colors (0 - 15, eventually including
 * white and black) are not to be overwritten by the 3D-viewer image
 * drawing routine now.
 * The initial brightness level has been increased from 3 to 4 (43% to 57%).
 *
 * Revision 1.14  1996/04/07  20:04:56  kans
 * getlayertop3d needed (void) as prototype
 *
 * Revision 1.13  1996/04/04  18:58:39  vakatov
 * Slight corrections to the camera storing/restoring algorithm and
 * the relevant data to avoid the object/camera position mismatching --
 * while setting the camera attributes stored during the work with another
 * picture.
 *
 * Revision 1.12  1996/04/03  16:03:33  vakatov
 * Now ZoomAll() can keep the camera angle (ZOOM_KEEP_ANGLE method)
 *
 * Revision 1.11  1996/04/02  22:30:02  vakatov
 * AttachZoomPic3D() removed; now previously stored camera position
 * is to be restored within AttachPicture3D automatically, as soon as
 * valid (non-NULL and initialized) "camera" argument is provided
 *
 * Revision 1.10  1996/04/01  23:07:03  vakatov
 * Now one can use prevously restored camera position
 * in the "AttachZoomPic3D(...,TRUE)"
 *
 * Revision 1.8  1996/03/30  23:48:02  hogue
 * removed unused variables, fixed a return value.
 *
 * Revision 1.7  1996/03/30  23:24:47  hogue
 * Cast MemCpy and Memset, changed camera to Int4, fixed nonportable uses of MemCpy, Added an AttachZoomPic .  Cameras are still not always saved and restored correctly...
 *
*
*
* ==========================================================================
*/
#include <math.h>

#ifndef _VIBRANT_
#include <vibrant.h>
#endif

#ifndef _VIEWER3D_
#include <viewer3d.h>
#endif

#ifndef _PPICT3D_
#include <ppict3d.h>
#endif

#ifndef _IMAGE_
#include <image.h>
#endif

/*****************************************************************************
*
*   DEFINES
*
*****************************************************************************/
#define IMAGE_READY  0x1
#define VIEWER_QROT  0x2
#define ZOOM_PAN     0x10
#define ZOOM_PAN_P   0x20
#define ZOOM_WIN     0x40
#define ZOOM_WIN_P   0x80
#define ZOOM_QROT_P  0x100
#define ZOOM_MASK    0xFF0

#define CAMERA_KEY   8071967
#define FIND_BLOCK_SIZE 8

/*****************************************************************************
*
*   TYPEDEFS
*
*****************************************************************************/
typedef struct Nlm_pviewer3d {
  double           dAOrt[3][3];
  double           dC[3];
  double           dscale;
  Int4             a[3][3];
  Int4             c[3];
  Int4             scale;
  Int4             zmax;
  Nlm_PPict3DPtr   pic;
  Image            image;
  Nlm_V3DClckProc  click;
  Nlm_V3DClckProc  drag;
  Nlm_V3DClckProc  release;
  Boolean          layerTable[256];
  Int4             p_version;
  Int4             g_version;
  Int4             g_versionPrev;
  Int4             c_version;
  Int4             c_versionPrev;
  Int4             colorOffset;
  Int4             colorStep;
  Int4             colorScale;
  PaneL            viewer;
  WindoW           parentWindow;
  Handle           zBuffer;
  Uint4            fondPrim[32];
  Uint2            nPrimFound;
  Int2             width;
  Int2             height;
  Int2             status;
  PoinT            p1;
  PoinT            p2;
  Uint1            perspectN;
  Uint1            brighN;
  Uint1            colorPicN;
  Uint1            colorHLR, colorHLG, colorHLB;
  Uint1            colorFon;
  Uint1            adjust;
  Uint1            perspect;
  Uint1            colorFonR, colorFonG, colorFonB;
} PViewer3D, PNTR PViewer3DPtr;

typedef struct
{
  Uint1  perspectN;
  Uint1  brighN;
  Uint1  colorPicN;
  Uint1  colorHLR, colorHLG, colorHLB;
  Uint1  colorFon;
  Uint1  adjust;
  Int4   scale;
  Uint1  colorFonR, colorFonG, colorFonB, dummy_align;
}  PCamera3DUnion;

/*****************************************************************************
*
*   GLOBAL VARIABLE
*
*****************************************************************************/
Uint1  Nlm_Isqrt[129*129];
extern Nlm_Context3D Nlm_stCon;

/*****************************************************************************
*
*   STATIC VARIABLE
*
*****************************************************************************/
static CharPtr      viewerClass = "Viewer3D";
static Boolean      isqrtInit = FALSE;
static Uint1        perspectAll[6] = { 8,6,4,3,2,1 };
static Int2         brighAll[7]    = { -150,-100,-50,0,50,100,150 };
static Uint1        colorPicAll[6] = { 2,3,4,6,8,255 };
static Uint1	    currlayer = 0;
static ButtoN       V3D_bPLAY;
static Uint1        toplayer = 0;
/*****************************************************************************
*
*   FUNCTIONS
*
*****************************************************************************/

static void BgColorProc(PViewer3DPtr p_viewer);


static void WorldToScreen ( PViewer3DPtr p, double x, double y, double z,
                            double * xS, double * yS, double * zS )
{
  register PViewer3DPtr pV;

  pV=p;
  *xS = pV->dC[0] + ( x * pV->dAOrt[0][0] +
                      y * pV->dAOrt[0][1] + 
                      z * pV->dAOrt[0][2] ) / pV->dscale;
  *yS = pV->dC[1] + ( x * pV->dAOrt[1][0] +
                      y * pV->dAOrt[1][1] + 
                      z * pV->dAOrt[1][2] ) / pV->dscale;
  *zS = pV->dC[2] + ( x * pV->dAOrt[2][0] +
                      y * pV->dAOrt[2][1] + 
                      z * pV->dAOrt[2][2] ) / pV->dscale;
}

static void RotateViewer ( PViewer3DPtr p, double PNTR mRot )
{
  register PViewer3DPtr pV;
  double   mRes[3][3];
  double   m[3][3];
  Int2     i,j;

  pV = p;
  MemCpy ( (void*)&(m[0][0]),(void*) mRot, (size_t)9*sizeof(double) );
  for ( j=0; j<3; j++ ){
    for ( i=0; i<3; i++ ){
      mRes[i][j] = m[i][0] * pV->dAOrt[0][j] + 
                   m[i][1] * pV->dAOrt[1][j] + 
                   m[i][2] * pV->dAOrt[2][j];
    }
  }
  MemCpy ( (void*)&(pV->dAOrt[0][0]), (void*)&(mRes[0][0]),(size_t) 9*sizeof(double) );
  pV->g_version++;
}

static void MakeOffsets ( PViewer3DPtr p,
                          double xWorld, double yWorld, double zWorld,
                          double xScreen, double yScreen, double zScreen )
{
  register PViewer3DPtr pV;

  pV = p;
  pV->dC[0] = xScreen - 
             ( xWorld * pV->dAOrt[0][0] +
               yWorld * pV->dAOrt[0][1] + 
               zWorld * pV->dAOrt[0][2] ) / pV->dscale;
  pV->dC[1] = yScreen - 
             ( xWorld * pV->dAOrt[1][0] +
               yWorld * pV->dAOrt[1][1] + 
               zWorld * pV->dAOrt[1][2] ) / pV->dscale;
  pV->dC[2] = zScreen - 
             ( xWorld * pV->dAOrt[2][0] +
               yWorld * pV->dAOrt[2][1] + 
               zWorld * pV->dAOrt[2][2] ) / pV->dscale;
  pV->g_version++;
}


static void MakeFloatMat ( PViewer3DPtr p )
{
  register Int2 i;
  register double PNTR dPtr;
  Int4            PNTR iPtr;

  p->dscale = (double) p->scale;

  dPtr = &(p->dAOrt[0][0]);
  iPtr = &(p->a[0][0]);
  for ( i=0; i<9; i++ )
    {
      *dPtr++ = (iPtr != 0)  ? (p->dscale / (double)*iPtr) : (p->dscale * 4);
      iPtr++;
    }
  dPtr = &(p->dC[0]);
  iPtr = &(p->c[0]);
  for ( i=0; i<3; i++ ) *dPtr++ = (double)(*iPtr++);
}


static void MakeIntMat ( PViewer3DPtr p )
{
  register Int2 i;
  register double PNTR dPtr;
  Int4            PNTR iPtr;
  double  tmpDoub;

  dPtr = &(p->dAOrt[0][0]);
  iPtr = &(p->a[0][0]);
  for ( i=0; i<9; i++ ){
    tmpDoub = *dPtr;
    if ( (tmpDoub < 1.0e-100) && (tmpDoub > -1.0e-100) ){
      *iPtr = INT4_MAX;
    } else {
      tmpDoub = p->dscale / tmpDoub;
      if ( tmpDoub > (double)INT4_MAX ) *iPtr = INT4_MAX;
      else if ( tmpDoub < (double)INT4_MIN ) *iPtr = INT4_MIN;
      else *iPtr = (Int4)tmpDoub;
    }
    dPtr++;
    iPtr++;
  }
  dPtr = &(p->dC[0]);
  iPtr = &(p->c[0]);
  for ( i=0; i<3; i++ ) *iPtr++ = (Int4)(*dPtr++);
  p->scale = (Int4)p->dscale;
}

static void SetUpContext ( PViewer3DPtr v )
{
  register PViewer3DPtr ax;

  ax=v;
  Nlm_stCon.xmin = (Int4)0;
  Nlm_stCon.xmax = (Int4)ax->width-1;
  Nlm_stCon.ymin = (Int4)0;
  Nlm_stCon.ymax = (Int4)ax->height-1;
  Nlm_stCon.zmax = ax->zmax;
  Nlm_stCon.pic = (Nlm_Picture3D)ax->pic;
  Nlm_stCon.width = (Int4)ax->width;
  Nlm_stCon.height = (Int4)ax->height;
  MemCpy ((void*) &(Nlm_stCon.layerTable[0]),
           (void*)&(ax->layerTable[0]), (size_t)256*sizeof(Boolean) );
  MemCpy ( (void*)&(Nlm_stCon.a[0][0]),
           (void*)&(ax->a[0][0]), (size_t)9*sizeof(Int4) );
  MemCpy ( (void*)&(Nlm_stCon.c[0]),
           (void*)&(ax->c[0]),(size_t) 3*sizeof(Int4) );
  Nlm_stCon.scale = ax->scale;
  Nlm_stCon.colorOffset = ax->colorOffset;
  Nlm_stCon.colorStep = ax->colorStep;
  Nlm_stCon.colorScale = ax->colorScale;
  Nlm_stCon.perspect = ax->perspect;
  Nlm_stCon.zmaxPersp = Nlm_stCon.perspect * Nlm_stCon.zmax;
  Nlm_stCon.zmaxPersp1 = Nlm_stCon.zmaxPersp + Nlm_stCon.zmax;
  Nlm_stCon.colorHL = (Uint1)ax->pic->totalColors;
}


typedef enum
{
  ZOOM_ALL,
  ZOOM_KEEP_ANGLE,
  ZOOM_KEEP_ALL
}
Nlm_enumZoom3D;


static void ZoomAll(PViewer3DPtr p, Nlm_enumZoom3D method)
{
  register PViewer3DPtr ax = p;
  double   d;
  double   scale;
  Spher3D  sph;
  Int4     diametr;
  
  if ( !Nlm_GetSegSphere3D((Nlm_Picture3D)ax->pic, NULL, &sph) )
    return;

  diametr = sph.radius << 1;

  if (method == ZOOM_ALL)
    {
      ax->dAOrt[0][0] = ax->dAOrt[1][1] = ax->dAOrt[2][2] = (double)1;
      ax->dAOrt[0][1] = ax->dAOrt[0][2] =
      ax->dAOrt[1][0] = ax->dAOrt[1][2] =
      ax->dAOrt[2][0] = ax->dAOrt[2][1] = (double)0;
    }

  if (method != ZOOM_KEEP_ALL)
    {
      scale = (double)diametr / (double)ax->width;
      d = (double)diametr / (double)ax->height;
      if ( scale < d ) scale = d;
      if ( scale < 10000.0 ) scale = 10000.0;
      ax->dscale = (scale*2.0*(double)ax->perspect) / 
	(2.0*(double)ax->perspect + 1.0);
    }

  ax->zmax = (Int4)((double)diametr / ax->dscale + 0.5);

  if (method != ZOOM_KEEP_ALL)
    {
      MakeOffsets ( ax, (double)sph.x, (double)sph.y, (double)sph.z,
		    (double)(ax->width>>1),
		    (double)(ax->height>>1),
		    (double)((ax->zmax>>1) + 1));
    }
  else
    {
      ax->dC[2] = (double)((ax->zmax>>1) + 1) - 
	( (double)sph.x * ax->dAOrt[2][0] +
	  (double)sph.y * ax->dAOrt[2][1] + 
	  (double)sph.z * ax->dAOrt[2][2] ) / ax->dscale;
    }

  MakeIntMat (ax);

  ax->g_version++;
}


static void ZoomPan ( PViewer3DPtr p, Int2 dx, Int2 dy )
{
  register PViewer3DPtr ax;
  
  ax = p;
  ax->dC[0] += (double)dx;
  ax->dC[1] -= (double)dy;
  MakeIntMat (ax);
  ax->g_version++;
}


static void ZoomWin ( PViewer3DPtr p, Int2 x1, Int2 y1, 
                      Int2 x2, Int2 y2 )
{
  register PViewer3DPtr ax = p;
  double   scale, minscale;
  double   d;
  Spher3D  sph;
  int      width, height;
  RecT     r;

  if ( !Nlm_GetSegSphere3D((Nlm_Picture3D)ax->pic, NULL, &sph) )
    return;

  minscale = (double)sph.radius / 32000.0 / ax->dscale;

  ObjectRect ( ax->viewer, &r );
  x1 -= r.left;  x2 -= r.left;
  y1 -= r.top;   y2 -= r.top;
  width = x2 - x1;
  if ( width < 0 ) { 
    width = -width; x1 = x2;
  }
  y1 = (Int2)(p->height - y1);
  y2 = (Int2)(p->height - y2);
  height = y2 - y1;
  if ( height < 0 ) { 
    height = -height; y1 = y2;
  }
  scale = (double)width / (double)ax->width;
  d = (double)height / (double)ax->height;
  if ( d > scale ) scale = d;
  if ( scale < minscale ) scale = minscale;
  ax->dC[0] = (ax->dC[0] - (double)(x1+width/2))/scale + (double)(ax->width/2);
  ax->dC[1] = (ax->dC[1] - (double)(y1+height/2))/scale + (double)(ax->height/2);
  ax->dscale *= scale;
  ax->zmax = (Int4)((double)sph.radius * 2.0 / ax->dscale + 0.5);
  ax->dC[2] = (double)((ax->zmax>>1) + 1) - 
             ( (double)sph.x * ax->dAOrt[2][0] +
               (double)sph.y * ax->dAOrt[2][1] + 
               (double)sph.z * ax->dAOrt[2][2] ) / ax->dscale;
  MakeIntMat (ax);
  ax->g_version++;
}


static void rotateXY ( PViewer3DPtr p, Int2 dAngle, Boolean rotX )
{
  register PViewer3DPtr ax = p;
  double   dx, dy, dz;
  double   d;
  double   ssin;
  double   ccos;
  double   a[3][3];
  Spher3D  sph;
  Int2     i,j;
  
  if ( !Nlm_GetSegSphere3D((Nlm_Picture3D)ax->pic, NULL, &sph) )
    return;

  WorldToScreen ( ax, (double)sph.x, (double)sph.y, (double)sph.z, 
                  &dx, &dy, &dz );
  d = (double)dAngle * NCBIMATH_PI / 180.0;
  ssin = sin ( d );
  ccos = cos ( d );
  for( i=0; i<3; i++ )
    for ( j=0; j<3; j++ ) a[i][j] = 0.0;
  if ( rotX ) {
    a[0][0] = 1;
    a[1][1] = ccos; a[1][2] = ssin;
    a[2][1] = -ssin; a[2][2] = ccos;
  } else {
    a[1][1] = 1;
    a[0][0] = ccos; a[0][2] = ssin;
    a[2][0] = -ssin; a[2][2] = ccos;
  }
  RotateViewer ( ax, &(a[0][0]) );
  MakeOffsets ( ax, (double)sph.x, (double)sph.y, (double)sph.z,
                dx, dy, dz );
  MakeIntMat (ax);
}

static void RedrawZStatus ( PViewer3DPtr viewer, Boolean setPort )
{
  register PViewer3DPtr pV;
  WindoW   tempPort;
  RecT     r;

  pV=viewer;
  if ( (pV->status & ZOOM_PAN_P) || (pV->status & ZOOM_WIN_P) ){
    if ( setPort ){
        tempPort = SavePort (pV->viewer);
        Select (pV->viewer);
    }
#ifdef WIN_MOTIF
    SetColor ( 0xf1 );
#endif
    InvertMode();
    if ( pV->status & ZOOM_PAN_P ){
      DrawLine (pV->p1, pV->p2);
    } else {
      LoadRect ( &r, pV->p1.x, pV->p1.y, pV->p2.x, pV->p2.y );
      FrameRect (&r);
    }
    CopyMode();
    if ( setPort ){
      RestorePort (tempPort);
    }
  }
}

static void CancelZStatus ( PViewer3DPtr viewer, Boolean setPort )
{
  RedrawZStatus ( viewer, setPort );
  viewer->status &= ~ZOOM_MASK;
}

static void DrawViewer3D (PaneL viewer)
{
  register PViewer3DPtr   pV;
  register Nlm_PPict3DPtr pP;
  PViewer3DPtr  extra;
  Uint1         i, k, colorStep;
  Uint2         j;
  Int4          red, green, blue;
  Int4          colorPic;
  Int4          brigh,kbrigh;
  RecT          r;
  PoinT         p;
  Uint2         iwidth, iheight;

  ObjectRect ( viewer, &r );
  GetPanelExtra (viewer, &extra);
  if ( extra == NULL ) return;
  pV = extra;
  pP = pV->pic;
  if ( pP == NULL ) return;
  Nlm_stCon.image = LockPixMapImage ( pV->image );
  if ( Nlm_stCon.image == NULL ) {
    Nlm_DiagPutRecord ( DA_ERROR, viewerClass, "DrawViewer3D",
                        "Can not lock memory block" );
    return;
  }
#ifdef WIN_MOTIF
  p.x = 2;
  p.y = 2;
#else
  p.x = (Int2)(r.left + 2);
  p.y = (Int2)(r.top  + 2);
#endif
  if ( (pP->version != pV->p_version)||
       (pV->c_version != pV->c_versionPrev) ){
    pV->colorOffset = 33;
    colorStep = (Uint1)((256 - pV->colorOffset) / (pP->totalColors + 1));
    if (colorStep > 32)
      colorStep = 32;
    pV->colorStep = colorStep;
    i = (Uint1)pV->colorOffset;
    colorPic = (Int4)colorPicAll[pV->colorPicN];
    brigh = (Int4)brighAll[pV->brighN];
    for (j=0; j<pP->totalColors; j++)  {
      red = (Int4)pP->colorR[j]; 
      green = (Int4)pP->colorG[j];
      blue = (Int4)pP->colorB[j];
      red = red + 2*(255-red)/colorPic;
      green = green + 2*(255-green)/colorPic;
      blue = blue + 2*(255-blue)/colorPic;
      for ( k=1; k<=colorStep; k++){
        if ( brigh >= 0 ){
          kbrigh = ((256-brigh)*k + brigh*colorStep)/colorStep;
        } else {
          kbrigh = ((256+brigh)*k)/colorStep;
        }
        SetColorImage (pV->image, i++, (Uint1)((red*kbrigh)>>8),
                                       (Uint1)((green*kbrigh)>>8),
                                       (Uint1)((blue*kbrigh)>>8));
      }
    }
    red = (Int4)pV->colorHLR; 
    green = (Int4)pV->colorHLG;
    blue = (Int4)pV->colorHLB;
    for ( k=1; k<=colorStep; k++){
      if ( brigh >= 0 ){
        kbrigh = ((256-brigh)*k + brigh*colorStep)/colorStep;
      } else {
        kbrigh = ((256+brigh)*k)/colorStep;
      }
      SetColorImage (pV->image, i++, (Uint1)((red*kbrigh)>>8),
                    (Uint1)((green*kbrigh)>>8),
                    (Uint1)((blue*kbrigh)>>8));
    }
    SetColorImage(pV->image, pV->colorFon,
                  pV->colorFonR, pV->colorFonG, pV->colorFonB);
    ImageSetPalette ( pV->image, pV->parentWindow );
    pV->c_versionPrev = pV->c_version;
    pV->p_version = pP->version;
    pV->g_version++;
  }
  if ( pV->g_version != pV->g_versionPrev ){
    Nlm_stCon.zBuffer = HandLock( pV->zBuffer );
    if ( Nlm_stCon.zBuffer == NULL ) {
      Nlm_DiagPutRecord ( DA_ERROR, viewerClass, "DrawViewer3D",
                          "Can not lock memory block" );
      return;
    }
    pV->colorScale = (pV->zmax<<8)/pV->colorStep;
    GetImageSize ( pV->image, &iwidth, &iheight );
    MemSet ( (void*)Nlm_stCon.image, pV->colorFon, iwidth*iheight );
    MemSet ( (void*)Nlm_stCon.zBuffer, 0, iwidth*iheight*2 );
    SetUpContext ( pV );
    if ( Nlm_UpdateGver( pP ) )
      pP->seg.base.fTable->draw ((Nlm_VoidPtr)&(pP->seg));
    ImageModified ( pV->image );
    pV->status |= IMAGE_READY;
    pV->g_versionPrev = pV->g_version;
    HandUnlock ( pV->zBuffer );
  }
  ImageShow ( pV->image, p );
  UnlockPixMapImage ( pV->image );
  RedrawZStatus ( pV, FALSE );
}

static void ResetViewerProc (PaneL viewer)
{
  register PViewer3DPtr ax;
  PViewer3DPtr          extra;
  Int2                  i;

  Nlm_DiagReset ();
  GetPanelExtra (viewer, &extra);
  ax = extra;
  if ( ax != NULL ) {
    if ( ax->image != NULL ) DeleteImage ( ax->image );
    if ( ax->zBuffer != NULL ) HandFree ( ax->zBuffer );
    MemSet ( (void*)ax, 0, sizeof(PViewer3D) );
    for ( i=0; i<256; i++ ){
      ax->layerTable[i] = TRUE;
    }
    ax->viewer = viewer;
  }
}

static void ViewerClickProc (PaneL viewer, PoinT pt)
{
  register PViewer3DPtr ax;
  PViewer3DPtr          extra;

  GetPanelExtra (viewer, &extra);
  ax = extra;
  if ((ax != NULL) && (ax->status & IMAGE_READY) ) {
    switch ( ax->status & ZOOM_MASK ){
      case ZOOM_PAN:
        ax->p1 = pt;
        ax->p2 = pt;
        ax->status &= ~ZOOM_MASK;
        ax->status |= ZOOM_PAN_P;
        RedrawZStatus ( ax, FALSE );
        break;
      case ZOOM_WIN:
        ax->p1 = pt;
        ax->p2 = pt;
        ax->status &= ~ZOOM_MASK;
        ax->status |= ZOOM_WIN_P;
        RedrawZStatus ( ax, FALSE );
        break;
      default:
        if ( ax->status & VIEWER_QROT ) {
          ax->p1 = pt;
          ax->status |= ZOOM_QROT_P;
        } else if (ax->click != NULL) {
          ax->click ( (Nlm_Viewer3D)viewer, pt );
        }
    }

    if ( Nlm_dblClick )
      BgColorProc( (PViewer3DPtr)ax );
  }
}

static void QRotate ( PaneL viewer, PViewer3DPtr pv, Int4 dxAn, Int4 dyAn )
{
  register PViewer3DPtr ax = pv;
  double   dx, dy, dz;
  double   d;
  double   ssin;
  double   ccos;
  double   a[3][3];
  Spher3D  sph;
  BaR      sbx, sby;
  int      i,j;
  Int2     valx, valy;

  if ( !Nlm_GetSegSphere3D((Nlm_Picture3D)ax->pic, NULL, &sph) )
    return;

  WorldToScreen ( ax, (double)sph.x, (double)sph.y, (double)sph.z, 
                  &dx, &dy, &dz );

  d = (double)dxAn/(double)ax->width;
  sbx = GetSlateHScrollBar ((SlatE) viewer);
  valx = (Int2)((int)(GetValue(sbx) + d * 180 + 360) % 360);
  d *= NCBIMATH_PI;
  ssin = sin ( d );
  ccos = cos ( d );
  for ( i=0; i<3; i++ )
  for ( j=0; j<3; j++ )
    a[i][j] = 0.0;
  a[1][1] = 1;
  a[0][0] =  ccos;  a[0][2] = ssin;
  a[2][0] = -ssin;  a[2][2] = ccos;
  RotateViewer ( ax, &(a[0][0]) );
  
  d = -(double)dyAn/(double)ax->height;
  sby = GetSlateVScrollBar ((SlatE)viewer);
  valy = (Int2)((int)(GetValue(sby) - d * 180 + 360) % 360);
  d *= NCBIMATH_PI;
  ssin = sin ( d );
  ccos = cos ( d );
  for ( i=0; i<3; i++ )
  for ( j=0; j<3; j++ )
    a[i][j] = 0.0;
  a[0][0] = 1;
  a[1][1] =  ccos; a[1][2] = ssin;
  a[2][1] = -ssin; a[2][2] = ccos;
  RotateViewer ( ax, &(a[0][0]) );

  MakeOffsets ( ax, (double)sph.x, (double)sph.y, (double)sph.z,
                dx, dy, dz );
  MakeIntMat (ax);
  DrawViewer3D (viewer);
  ResetClip ();
  CorrectBarValue ( sbx, valx );
  CorrectBarValue ( sby, valy );
}

static void ViewerDragProc (PaneL viewer, PoinT pt)
{
  register PViewer3DPtr ax;
  PViewer3DPtr          extra;

  GetPanelExtra (viewer, &extra);
  ax = extra;
  if ((ax != NULL) && (ax->status & IMAGE_READY) ) {
    switch ( ax->status & ZOOM_MASK ){
      case ZOOM_WIN_P:
      case ZOOM_PAN_P:
        RedrawZStatus ( ax, FALSE );
        ax->p2 = pt;
        RedrawZStatus ( ax, FALSE );
        break;

      case ZOOM_QROT_P:
        QRotate ( viewer, ax, 
                  (Int4)(pt.x - ax->p1.x), (Int4)(pt.y - ax->p1.y));
        ax->p1 = pt;
        break;
      default:
        if (ax->drag != NULL) {
          ax->drag ( (Nlm_Viewer3D)viewer, pt );
        }
    }
  }
}

static void ViewerReleaseProc (PaneL viewer, PoinT pt)
{
  register PViewer3DPtr ax;
  PViewer3DPtr  extra;

  GetPanelExtra (viewer, &extra);
  ax = extra;
  if ((ax != NULL) && (ax->status & IMAGE_READY) ) {
    switch ( ax->status & ZOOM_MASK ){
      case ZOOM_PAN_P:
        CancelZStatus ( ax, FALSE );
        ax->p2 = pt;
        if ( (ax->p1.x != ax->p2.x) || (ax->p1.y != ax->p2.y) ){
          ZoomPan (ax, (Int2)(ax->p2.x - ax->p1.x), (Int2)(ax->p2.y - ax->p1.y));
          DrawViewer3D( viewer );
        }
        break;
      case ZOOM_WIN_P:
        CancelZStatus ( ax, FALSE );
        ax->p2 = pt;
        if ( (ax->p1.x != ax->p2.x) && (ax->p1.y != ax->p2.y) ){
          ZoomWin (ax, ax->p1.x, ax->p1.y, ax->p2.x, ax->p2.y);
          DrawViewer3D ( viewer );
        }
        break;
      case ZOOM_QROT_P:
        CancelZStatus (ax, FALSE);
        break;
      default:
        if (ax->release != NULL) {
          ax->release ((Nlm_Viewer3D)viewer, pt);
        }
    }
  }
}


static void ViewerVScrollProc (BaR sb, SlatE viewer, Int2 newval, Int2 oldval)
{
  register PViewer3DPtr ax;
  PViewer3DPtr  extra;
  WindoW        tempPort;

  if (viewer != NULL && oldval != newval) {
    if (Visible (viewer) && AllParentsVisible (viewer) ) {
      GetPanelExtra ((PaneL)viewer, &extra);
      if ( extra->status & IMAGE_READY ){
        ax = extra;
        rotateXY (ax, (Int2)(oldval - newval), TRUE);
        tempPort = CurrentWindow ();
        UseWindow (ParentWindow (viewer));
        Select (viewer);
        CancelZStatus ( ax, FALSE );
        DrawViewer3D ( (PaneL)viewer );
        UseWindow (tempPort);
        Nlm_DiagReset ();
      }
    }
  }
}

static void ViewerHScrollProc (BaR sb, SlatE viewer, Int2 newval, Int2 oldval)
{
  register PViewer3DPtr ax;
  PViewer3DPtr  extra;
  WindoW        tempPort;

  if (viewer != NULL && oldval != newval) {
    if (Visible (viewer) && AllParentsVisible (viewer)) {
      GetPanelExtra ((PaneL)viewer, &extra);
      if ( extra->status & IMAGE_READY ){
        ax = extra;
        rotateXY (ax, (Int2)(newval - oldval), FALSE );
        tempPort = CurrentWindow ();
        UseWindow (ParentWindow (viewer));
        Select (viewer);
        CancelZStatus ( ax, FALSE );
        DrawViewer3D ( (PaneL)viewer );
        UseWindow (tempPort);
        Nlm_DiagReset ();
      }
    }
  }
}

void Nlm_SetQRotate3D ( Pointer v3Dextra, Boolean onoff )
{
  register PViewer3DPtr v;

  v = v3Dextra;
  if ( (v != NULL) && (v->status & IMAGE_READY) )
    CancelZStatus ( v, TRUE );

  if ( onoff )
    v->status |= VIEWER_QROT;
  else
    v->status &= ~VIEWER_QROT;
}

Boolean Nlm_GetQRotate3D( Pointer v3Dextra)
{
  register PViewer3DPtr v;

  v = v3Dextra;
  if (!v) return FALSE;
  if (v->status & VIEWER_QROT) return TRUE;
  return FALSE;

}

static void QRotateProc ( ButtoN b )
{
  Nlm_SetQRotate3D ( (Pointer)GetObjectExtra(b), GetStatus(b) );
}

static void PlusMinus ( PViewer3DPtr viewer, Boolean plus )
{
  register PViewer3DPtr v;
  double                dx,dy,dz;
  Spher3D               sph;
  WindoW                tmpPort;
  Int2                  needRedraw = 0;

  v = viewer;
  switch ( v->adjust ){
    case ADJUST3D_PERSPECT:
      if ( plus ) {
        if ( v->perspectN < 5 ) {
          v->perspectN++; v->g_version++;
          needRedraw = 1;
        }
      } else {
        if ( v->perspectN > 0 ) {
          v->perspectN--; v->g_version++;
          needRedraw = 1;
        }
      }
      if (needRedraw  &&
          Nlm_GetSegSphere3D((Nlm_Picture3D)v->pic, NULL, &sph) )
        {
          WorldToScreen ( v, (double)sph.x, (double)sph.y, (double)sph.z, 
                          &dx, &dy, &dz );
          v->dscale = (v->dscale * (2.0*(double)v->perspect + 1.0)) / 
            (2.0*(double)v->perspect);
          v->perspect = perspectAll[v->perspectN];
          v->dscale = (v->dscale * 2.0*(double)v->perspect) / 
            (2.0*(double)v->perspect + 1.0);
          v->zmax = (Int4)((double)(sph.radius<<1) / v->dscale + 0.5);
          MakeOffsets ( v, (double)sph.x, (double)sph.y, (double)sph.z,
                        dx, dy, (double)((v->zmax>>1) + 1) );
          MakeIntMat (v);
        }
      break;
    case ADJUST3D_BRIGH:
      if ( plus ) {
        if ( v->brighN < 6 ) {
          v->brighN++; v->c_version++;
          needRedraw = 1;
        }
      } else {
        if ( v->brighN > 0 ) {
          v->brighN--; v->c_version++;
          needRedraw = 1;
        }
      }
      break;
    case ADJUST3D_COLOR:
      if ( plus ) {
        if ( v->colorPicN < 5 ) {
          v->colorPicN++; v->c_version++;
          needRedraw = 1;
        }
      } else {
        if ( v->colorPicN > 0 ) {
          v->colorPicN--; v->c_version++;
          needRedraw = 1;
        }
      }
  }
  if ( needRedraw ){
    tmpPort = SavePort (v->viewer);
    Select (v->viewer);
    CancelZStatus ( v, FALSE );
    DrawViewer3D ( v->viewer );
    RestorePort (tmpPort);
    Nlm_DiagReset ();
  }
}


static void BgColorProc(PViewer3DPtr p_viewer)
{
  if (p_viewer == NULL  ||  p_viewer->viewer == NULL)
    return;

    if ( ChooseColorDialog (&p_viewer->colorFonR,
                            &p_viewer->colorFonG,
                            &p_viewer->colorFonB,
                            0) )
      {
        Nlm_SetBackground3D((Nlm_Viewer3D)(p_viewer->viewer),
                            p_viewer->colorFonR,
                            p_viewer->colorFonG,
                            p_viewer->colorFonB);
      }
}


void Nlm_SetAdjust3D ( Pointer v3Dextra, Int2 ajust )
{
  register PViewer3DPtr v;

  v = (PViewer3DPtr)v3Dextra;
  if ( v!=NULL ) v->adjust = (Uint1)ajust;
}
 
static void AdjustProc ( GrouP g )
{
  Nlm_SetAdjust3D ( (Pointer)GetObjectExtra(g), GetValue(g) );
}

void Nlm_AdjustPlus3D ( Pointer v3Dextra )
{
  register PViewer3DPtr v;

  v = (PViewer3DPtr)v3Dextra;
  if ( (v != NULL) && (v->status & IMAGE_READY) ) {
    PlusMinus ( v, TRUE );
  }
}

static void PlusProc ( ButtoN b )
{
  Nlm_AdjustPlus3D ( (Pointer)GetObjectExtra(b) );
}

void Nlm_AdjustMinus3D ( Pointer v3Dextra )
{
  register PViewer3DPtr v;

  v = (PViewer3DPtr)v3Dextra;
  if ( (v != NULL) && (v->status & IMAGE_READY) ) {
    PlusMinus ( v, FALSE );
  }
}

static void MinusProc ( ButtoN b )
{
  Nlm_AdjustMinus3D ( (Pointer)GetObjectExtra(b) );
}

void Nlm_ZoomAll3D ( Pointer v3Dextra )
{
  register PViewer3DPtr v;
  WindoW                tmpPort;

  v = (PViewer3DPtr)v3Dextra;
  if ( v != NULL ){
    if ( v->status & IMAGE_READY ) {
      if (Visible (v->viewer) && AllParentsVisible (v->viewer)) {
        ZoomAll (v, ZOOM_KEEP_ANGLE);
        tmpPort = SavePort (v->viewer);
        Select (v->viewer);
        CancelZStatus ( v, FALSE );
        DrawViewer3D ( v->viewer );
        RestorePort (tmpPort);
        Nlm_DiagReset ();
      }
    }
  }
}
 
static void ZoomAllProc ( ButtoN b )
{
  Nlm_ZoomAll3D ( (Pointer)GetObjectExtra(b) );
}

void Nlm_ZoomPan3D ( Pointer v3Dextra )
{
  register PViewer3DPtr v;

  v = (PViewer3DPtr)v3Dextra;
  if ( (v != NULL) && (v->status & IMAGE_READY) ) {
    CancelZStatus ( v, TRUE );
    v->status |= ZOOM_PAN;
  }
}

static void ZoomPanProc ( ButtoN b ) 
{
  Nlm_ZoomPan3D ( (Pointer)GetObjectExtra(b) );
}

void Nlm_ZoomWin3D ( Pointer v3Dextra )
{
  register PViewer3DPtr v;

  v = (PViewer3DPtr)v3Dextra;
  if ( (v != NULL) && (v->status & IMAGE_READY) ) {
    CancelZStatus ( v, TRUE );
    v->status |= ZOOM_WIN;
  }
}

static void ZoomWinProc ( ButtoN b ) 
{
  Nlm_ZoomWin3D ( (Pointer)GetObjectExtra(b) );
}

#ifndef WIN_MOTIF

void Nlm_CopyViewer3D ( Pointer v3Dextra )
{
  register PViewer3DPtr v;

  v = (PViewer3DPtr)v3Dextra;
  if ( (v != NULL) && (v->status & IMAGE_READY) ) {
    CancelZStatus ( v, TRUE );
    SaveImageClip ( v->image );
  }
}

static void CopyProc ( ButtoN b ) 
{
  Nlm_CopyViewer3D ( (Pointer)GetObjectExtra(b) );
}

void Nlm_PrintViewer3D ( Pointer v3Dextra )
{
  register PViewer3DPtr v;

  v = (PViewer3DPtr)v3Dextra;
  if ( (v != NULL) && (v->status & IMAGE_READY) ) {
    CancelZStatus ( v, TRUE );
    PrintImage ( v->image );
  }
}

static void PrintProc ( ButtoN b ) 
{
  Nlm_PrintViewer3D ( (Pointer)GetObjectExtra(b) );
}
#endif

void Nlm_SaveViewer3D ( Pointer v3Dextra )
{
  Char fname[PATH_MAX];
  PViewer3DPtr v = (PViewer3DPtr)v3Dextra;
  if (v == NULL  ||  (v->status & IMAGE_READY) == 0)
    return;

  if ( !GetOutputFileName(fname, sizeof(fname), "*.gif") )
    return;

  CancelZStatus(v, TRUE);
  SaveImageGIF(v->image, fname);
}

static void SaveProc ( ButtoN b ) 
{
  Nlm_SaveViewer3D ( (Pointer)GetObjectExtra(b) );
}

Nlm_Viewer3D Nlm_CreateViewer3D (Nlm_GrouP prnt, Uint2Ptr width, Uint2 height,
                                 Boolean vscroll, Boolean hscroll) 
{
  register PViewer3DPtr ax;
  PViewer3DPtr  vPtr;
  PaneL         viewer;
  SltScrlProc   hScrProc = NULL;
  SltScrlProc   vScrProc = NULL;
  Int4          i,j,k;
  BaR           sb;
  Int2          margin;

  Nlm_DiagReset ();
#ifdef WIN_MSWIN
#ifndef WIN32
  if ( ((Int4)(*width) * (Int4)height) > (Int4)0x7FFF ) {
      Nlm_DiagPutRecord ( DA_ERROR, viewerClass, "CreateViewer3D",
                          "Invalid size of viewer window");
      return NULL;
  }
#endif
#endif

  if ( isqrtInit == FALSE ){
    k = 0;
    for ( i=0; i<128; i++ ){
      for ( j=-i; j<=i; j++ ){
        Nlm_Isqrt[k++] = (Uint1)(sqrt ( (double)(i*i - j*j) ));
      }
    }
    isqrtInit = TRUE;
  }

  if (prnt == NULL  ||
      *width < (Uint2)(16 + (hscroll ?  Nlm_vScrollBarWidth  : 0))  ||
      height < (Uint2)(16 + (vscroll ?  Nlm_hScrollBarHeight : 0)))
    {
      Nlm_DiagPutRecord(DA_ERROR, viewerClass, "CreateViewer3D",
                        "Invalid argument" );
      return NULL;
    }
  if ( hscroll )  *width -= Nlm_vScrollBarWidth ;
  if ( vscroll )  height -= Nlm_hScrollBarHeight;

  ax = (PViewer3DPtr)MemNew ( sizeof(PViewer3D) );
  if ( ax == NULL ) {
    Nlm_DiagPutRecord ( DA_ERROR, viewerClass, "CreateViewer3D",
                        "Can not allocate memory block" );
    return NULL;
  }


  ax->parentWindow = ParentWindow ( (Handle)prnt );
  ax->image = CreateImage();
  if ( ax->image == NULL ) {
    MemFree ( ax ); return NULL;
  }
  if ( AllocateImage ( ax->image, width, height, 32, 256 ) != TRUE ){
    DeleteImage ( ax->image );
    MemFree ( ax ); return NULL;
  }
  ax->width = *width;
  ax->height = height;
  ax->colorHLR = ax->colorHLG = ax->colorHLB = (Uint1)255;
  ax->colorFon = ImageGetBlack( ax-> image );
  ax->colorFonR = ax->colorFonG = ax->colorFonB = (Uint1)0;
  ax->zBuffer = HandNew ( ((Int4)(*width) * (Int4)height) << 1 );
  if ( ax->zBuffer == NULL ) {
    Nlm_DiagPutRecord ( DA_ERROR, viewerClass, "CreateViewer3D",
                        "Can not allocate memory block" );
    DeleteImage ( ax->image );
    MemFree ( ax ); return NULL;
  }
  for ( i=0; i<256; i++ ){
    ax->layerTable[i] = TRUE;
  }
  ax->g_version++;
  ax->c_version++;
  margin = -3;
  if ( vscroll ) vScrProc = ViewerVScrollProc;
  if ( hscroll ) hScrProc = ViewerHScrollProc;
  viewer = AutonomousPanel (prnt, (Int2)(*width + margin), (Int2)(height + margin),
                            DrawViewer3D, vScrProc, hScrProc,
                            sizeof(BigScalar), ResetViewerProc, NULL);
  if ( viewer == NULL ) {
    Nlm_DiagPutRecord ( DA_ERROR, viewerClass, "CreateViewer3D",
                        "Can not create panel" );
    DeleteImage ( ax->image );
    HandFree ( ax->zBuffer );
    MemFree ( ax ); return NULL;
  }
  if ( hscroll ) {
    sb = GetSlateHScrollBar ((SlatE) viewer);
    CorrectBarValue(sb, 0);
    SetRange (sb, 10, 10, 360);
  }
  if ( vscroll ) {
    sb = GetSlateVScrollBar ((SlatE) viewer);
    CorrectBarValue(sb, 0);
    SetRange (sb, 10, 10, 360);
  }
  SetPanelClick (viewer, ViewerClickProc, ViewerDragProc, NULL, 
                 ViewerReleaseProc);
  ax->adjust = ADJUST3D_PERSPECT;
  ax->viewer = viewer;
  ax->status |= VIEWER_QROT;
  vPtr = ax;
  SetPanelExtra (viewer, &vPtr );
  return (Nlm_Viewer3D)viewer;
}


static void Nlm_AllLayerSetStates3D(PViewer3DPtr viewer, Nlm_Boolean state)
{
    register Uint1 level = 0;
    do
      {
        SetLayer3D((Nlm_Picture3D)viewer->pic, level, state);
        viewer->layerTable[level] = state;
      }
    while (level++ != 255);
}


static void Nlm_AllLayerSet3D(PViewer3DPtr v, Nlm_Boolean state)
{
  register Nlm_PPict3DPtr pP;
  Spher3D  sph;
  double   dx, dy, dz;
  
  if (v == NULL  ||  (pP = v->pic) == NULL)  return;

  Nlm_AllLayerSetStates3D(v, state);

  if ( Nlm_GetSegSphere3D((Nlm_Picture3D)pP, NULL, &sph) )
    {
      WorldToScreen (v,
                     (double)sph.x, (double)sph.y, (double)sph.z,
                     &dx, &dy, &dz );
      MakeOffsets   (v,
                     (double)sph.x, (double)sph.y, (double)sph.z,
                     dx,  dy,  dz );
      MakeIntMat( v ); 
    }

  RedrawViewer3D ( (Nlm_Viewer3D)v->viewer );
}


static void AllLayerOnProc( ButtoN b )
{
  Nlm_AllLayerSet3D((PViewer3DPtr)GetObjectExtra(b), TRUE);
}


static void Nlm_RewindLayer3D(PViewer3DPtr v)
{
  if (v == NULL  ||  v->pic == NULL)  return;

  Nlm_AllLayerSetStates3D(v, FALSE);
  
  SetLayer3D((Nlm_Picture3D)v->pic, 1, TRUE); 
  v->layerTable[0] = TRUE; 
  currlayer = 0;
  RedrawViewer3D( (Nlm_Viewer3D)v->viewer );
}


static void RewindLayerProc( ButtoN b )
{
  Nlm_RewindLayer3D( (PViewer3DPtr)GetObjectExtra( b ) );
}


void Nlm_PrevLayer3D (Pointer v3Dextra)
{
  register PViewer3DPtr v = (PViewer3DPtr)v3Dextra;

  if (v == NULL  ||  v->pic == NULL)  return;

  Nlm_AllLayerSetStates3D(v, FALSE);

  currlayer = (currlayer != 0) ?  (currlayer - 1) : toplayer;
  SetLayer3D((Nlm_Picture3D)v->pic, currlayer, TRUE); 
  v->layerTable[currlayer] = TRUE; 

  RedrawViewer3D( (Nlm_Viewer3D)v->viewer );
}


static void PrevLayerProc(ButtoN b)
{
  Nlm_PrevLayer3D( (Pointer)GetObjectExtra( b ) );
}


void Nlm_NextLayer3D (Pointer v3Dextra)
{
  register PViewer3DPtr v = (PViewer3DPtr)v3Dextra;

  if (v == NULL  ||  v->pic == NULL)  return;

  Nlm_AllLayerSetStates3D(v, FALSE);

  currlayer = (currlayer != toplayer) ?  (currlayer + 1) : 0;
  SetLayer3D((Nlm_Picture3D)v->pic, currlayer, TRUE); 
  v->layerTable[currlayer] = TRUE; 

  RedrawViewer3D ( (Nlm_Viewer3D)v->viewer );
}


static void NextLayerProc(ButtoN b)
{
  Nlm_PrevLayer3D ((Pointer)GetObjectExtra( b ) );
}


extern Boolean Nlm_IsPlaying3D(void)
{
#ifdef WIN_MAC
   return FALSE;
#else
   return GetStatus(V3D_bPLAY);
#endif
}


extern void  Nlm_StopPlaying3D  (void) 
{
  SetStatus(V3D_bPLAY, FALSE);
}

extern void Nlm_StartPlaying3D (Pointer v3Dextra )
{
#ifndef WIN_MAC
  SetStatus(V3D_bPLAY, TRUE);
  Nlm_PlayLayer3D(v3Dextra);
#endif
  return;
}


void Nlm_PlayLayer3D ( Pointer v3Dextra )
{
#ifdef WIN_MAC
  return;  /* not supported */
#else
  register PViewer3DPtr v = (PViewer3DPtr) v3Dextra;
  
  if (v == NULL  ||  v->pic == NULL)  return;

  Nlm_AllLayerSetStates3D(v, FALSE);
  
  SetLayer3D((Nlm_Picture3D)v->pic, currlayer, TRUE);  
  v->layerTable[currlayer] = TRUE; 

  while(GetStatus(V3D_bPLAY) == TRUE)
    {
      SetLayer3D((Nlm_Picture3D)v->pic, currlayer, FALSE); 
      v->layerTable[currlayer] = FALSE;  

      currlayer = (currlayer != toplayer) ? (currlayer + 1) : 0;
      SetLayer3D((Nlm_Picture3D)v->pic, currlayer, TRUE); 
      v->layerTable[currlayer] = TRUE; 

      RedrawViewer3D ((Nlm_Viewer3D) v->viewer );
      while ( Nlm_EventAvail() )
        Nlm_ProcessAnEvent ();
    }

#endif
}

       	
void Nlm_SetLayerTop3D(Uint1 top)
{
  toplayer = top;
}


Uint1 Nlm_GetLayerTop3D(void)
{
  return toplayer;
}	
	

static void PlayLayerProc( ButtoN b )
{
  Nlm_PlayLayer3D( (Pointer)GetObjectExtra( b ) );
}


GrouP Nlm_CreateAnimControl3D(Nlm_GrouP prnt, Boolean vertical, Boolean savebtn,
                              Nlm_V3DControlsPtr controls)
{
  GrouP mGroup;
  GrouP tmpGroup;
 
  mGroup = Nlm_CreateControl3DS (prnt, vertical, savebtn, controls);
  if (vertical)
    {
        StaticPrompt ( mGroup, "Animate:", 0, 0, Nlm_systemFont, 'r' );
        tmpGroup = HiddenGroup ( mGroup, 4, 1, NULL );
        controls->allButton = PushButton( tmpGroup, "*", AllLayerOnProc);
        controls->rewindButton = PushButton( tmpGroup, "|<", RewindLayerProc);
        controls->prevButton = PushButton(tmpGroup, "<", PrevLayerProc);
        controls->nextButton = PushButton(tmpGroup, ">", NextLayerProc);
#ifndef WIN_MAC
        V3D_bPLAY = controls->playButton = CheckBox(mGroup, "Go", PlayLayerProc);
#endif
    }
  else
    {
        controls->allButton = PushButton( mGroup, "*", AllLayerOnProc);
        controls->rewindButton = PushButton( mGroup, "|<", RewindLayerProc);
        controls->prevButton = PushButton(mGroup, "<", PrevLayerProc);
        controls->nextButton = PushButton(mGroup, ">", NextLayerProc);
#ifndef WIN_MAC
        V3D_bPLAY = controls->playButton = CheckBox(mGroup, "Go", PlayLayerProc);
#endif
    }
  return mGroup;
}




 



GrouP Nlm_CreateControl3D ( Nlm_GrouP prnt, Boolean vertical, 
                            Nlm_V3DControlsPtr controls)
{
   return Nlm_CreateControl3DS (prnt, vertical, TRUE, controls);
}



GrouP Nlm_CreateControl3DS ( Nlm_GrouP prnt, Boolean vertical, Boolean savebtn,
                            Nlm_V3DControlsPtr controls)
{
  GrouP mGroup;
  GrouP tmpGroup;
  GrouP tmptmpGroup;

  Nlm_DiagReset ();
  if ( vertical ){
    mGroup = HiddenGroup ( prnt, 1, 0, NULL );
    if ( mGroup != NULL ) {
#ifdef WIN_MOTIF
      SetGroupMargins ( mGroup, 0, 0 );
      SetGroupSpacing ( mGroup, 0, 0 );
#else
      SetGroupMargins ( mGroup, 1, 1 );
      SetGroupSpacing ( mGroup, 0, 0 );
#endif
      controls->rotate = CheckBox ( mGroup, "Rotate", QRotateProc );
      StaticPrompt ( mGroup, "Zoom:", 0, 0, Nlm_systemFont, 'c' );
      controls->zoomAll = PushButton ( mGroup, "All",  ZoomAllProc );
      controls->zoomPan = PushButton ( mGroup, "Move", ZoomPanProc );
      controls->zoomWin = PushButton ( mGroup, "Zoom", ZoomWinProc );
      StaticPrompt ( mGroup, "------", 0, 0, Nlm_systemFont, 'c' );
      controls->adjustButtons = tmpGroup = 
                                HiddenGroup ( mGroup, 1, 3, AdjustProc );
      RadioButton ( tmpGroup, "Pesp" );
      RadioButton ( tmpGroup, "Brigh" );
      RadioButton ( tmpGroup, "Color" );
      SetValue ( tmpGroup, 1 );
      tmpGroup = HiddenGroup ( mGroup, 2, 1, NULL );
      controls->minusButton = PushButton ( tmpGroup, "-",  MinusProc);
      controls->plusButton  = PushButton ( tmpGroup, "+",  PlusProc );
      StaticPrompt ( mGroup, "------", 0, 0, Nlm_systemFont, 'c' );

#ifndef WIN_MOTIF
      /*
      controls->copyButton  = PushButton ( mGroup,  "Copy",  CopyProc );
      controls->printButton = PushButton ( mGroup, "Print", PrintProc );
      */
#endif
      
     if (savebtn)
        controls->saveButton  = PushButton ( mGroup, "Save", SaveProc );
    }
  }  else {  /* horizontal*/

/* rearranged to use a single horizontal group */
    mGroup = HiddenGroup ( prnt, 0, 1, NULL );
    SetGroupSpacing ( mGroup, 6, 1 );
    if ( mGroup != NULL ) {
   
      tmpGroup = HiddenGroup ( mGroup, 3, 0, NULL );
      SetGroupSpacing ( tmpGroup, 1, 0 );
      controls->minusButton = PushButton ( tmpGroup, "-", MinusProc );
      controls->adjustButtons = tmptmpGroup = 
                                HiddenGroup ( tmpGroup, 3, 0, AdjustProc );
      RadioButton ( tmptmpGroup, "p" );
      RadioButton ( tmptmpGroup, "b" );
      RadioButton ( tmptmpGroup, "c" );
      SetValue ( tmptmpGroup, 1 );      
      controls->plusButton = PushButton ( tmpGroup, "+", PlusProc );
      
      if (savebtn) 
         controls->saveButton = PushButton ( mGroup, "Save", SaveProc );

      tmpGroup = HiddenGroup ( mGroup, 4, 0, NULL );
      controls->zoomAll = PushButton ( tmpGroup, "All",  ZoomAllProc );
      controls->zoomPan = PushButton ( tmpGroup, "Move", ZoomPanProc );
      controls->zoomWin = PushButton ( tmpGroup, "Zoom", ZoomWinProc );
      controls->rotate = CheckBox ( mGroup, "Turn", QRotateProc );
      SetStatus(controls->rotate, TRUE);
#ifndef WIN_MOTIF
      /*
      controls->copyButton  = PushButton ( mGroup, "Copy", CopyProc );
      controls->printButton = PushButton ( mGroup, "Print", PrintProc );
      */
#endif
    }
  }
  
  return mGroup;
}


Pointer Nlm_GetViewerExtra3D ( Nlm_Viewer3D viewer )
{
  Pointer extra;

  GetPanelExtra ((PaneL)viewer,&extra);
  return extra;
}


Nlm_Image Nlm_GetViewerImage3D(Nlm_Viewer3D viewer)
{
  PViewer3DPtr extra;
  GetPanelExtra((PaneL)viewer,&extra);

  return extra->image;
}


void Nlm_LinkControls3D ( Nlm_V3DControlsPtr controls, Pointer extra )
{
  if ( controls != NULL ){
    SetObjectExtra (controls->rotate,        (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->zoomAll,       (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->zoomPan,       (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->zoomWin,       (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->adjustButtons, (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->plusButton,    (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->minusButton,   (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->saveButton,    (Nlm_VoidPtr)extra, NULL);
#ifdef WIN_MSWIN
    /*
    SetObjectExtra (controls->copyButton,    (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->printButton,   (Nlm_VoidPtr)extra, NULL);
    */
#endif
  }
}


void Nlm_LinkAnimCtrls3D ( Nlm_V3DControlsPtr controls, Pointer extra )
{
   if ( controls != NULL) {
    Nlm_LinkControls3D(controls,  extra);
    SetObjectExtra (controls->allButton,    (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->rewindButton, (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->prevButton,   (Nlm_VoidPtr)extra, NULL);
    SetObjectExtra (controls->nextButton,   (Nlm_VoidPtr)extra, NULL);
#ifndef WIN_MAC
    SetObjectExtra (controls->playButton,   (Nlm_VoidPtr)extra, NULL);
#endif    
   }
}


void Nlm_DeleteViewer3D ( Nlm_Viewer3D viewer )
{
  PViewer3DPtr  extra;

  Nlm_DiagReset ();
  GetPanelExtra ((PaneL)viewer,&extra);
  Remove (viewer);
  if ( extra != NULL ) MemFree ( extra );
}

Boolean Nlm_AttachPicture3D ( Nlm_Viewer3D viewer, Picture3D picture, 
                              Nlm_Camera3DPtr camera )
{
  register PViewer3DPtr ax;
  PViewer3DPtr   extra;
  WindoW         tempPort;
  PCamera3DUnion camUnn;
  BaR            sb;
  RecT           r;
  Nlm_PPict3DPtr pP;
  

  Nlm_DiagReset ();
  GetPanelExtra ((PaneL)viewer, &extra);
  if ((ax = extra) == NULL)
    return FALSE;

  pP = ax->pic = (Nlm_PPict3DPtr)picture;

  sb = GetSlateHScrollBar ((SlatE)viewer);
  if ( sb != NULL ) {
    CorrectBarValue(sb, 0);
    SetRange (sb, 10, 10, 360);
  }
  sb = GetSlateVScrollBar ((SlatE)viewer);
  if ( sb != NULL ) {
    CorrectBarValue(sb, 0);
    SetRange (sb, 10, 10, 360);
  }

  ASSERT ( sizeof(ax->layerTable)  ==  sizeof(pP->layerTable) );
  MemCpy((void*)ax->layerTable, (void*)pP->layerTable,
         sizeof(ax->layerTable));
  SetLayer3D((Nlm_Picture3D)pP, 1, TRUE);
  ax->layerTable[0] = TRUE; 
  currlayer = 0;
  ASSERT ( CAMERA_SIZE_I4 * sizeof(Int4)  ==  sizeof(camera->dummy) );
  if (camera != NULL  &&  camera->dummy[0] == CAMERA_KEY)
    {
      ASSERT ( sizeof(ax->a) <=
	       sizeof(camera->dummy) - 1*sizeof(camera->dummy[0]) ); 
      MemCpy ((void*) &(ax->a[0][0]), (void*)&(camera->dummy[1] ),
	      sizeof(ax->a));

      ASSERT ( sizeof(ax->c) <=
	       sizeof(camera->dummy) - 10*sizeof(camera->dummy[0]) ); 
      MemCpy ((void*) &(ax->c[0]),    (void*)&(camera->dummy[10]),
	      sizeof(ax->c));

      ASSERT ( sizeof(camUnn) <=
	       sizeof(camera->dummy) - 13*sizeof(camera->dummy[0]) ); 
      MemCpy ((void*) &camUnn,        (void*)&(camera->dummy[13]),
	      sizeof(camUnn));

      ax->perspectN  = camUnn.perspectN;
      ax->perspect   = perspectAll[ax->perspectN];
      ax->brighN     = camUnn.brighN;
      ax->colorPicN  = camUnn.colorPicN;
      ax->colorHLR   = camUnn.colorHLR;
      ax->colorHLG   = camUnn.colorHLG;
      ax->colorHLB   = camUnn.colorHLB;
      ax->colorFon   = camUnn.colorFon;
      ax->colorFonR  = camUnn.colorFonR;
      ax->colorFonG  = camUnn.colorFonG;
      ax->colorFonB  = camUnn.colorFonB;
      ax->adjust     = camUnn.adjust;
      ax->scale      = camUnn.scale;
      MakeFloatMat( ax );
      SetUpContext ( ax );
      ZoomAll(ax, ZOOM_KEEP_ALL);
    }
  else
    {
      ax->perspectN = 2;
      ax->perspect = perspectAll[ax->perspectN];
      ax->brighN = 5;
      ax->colorPicN = 5;
      SetUpContext ( ax );
      ZoomAll(ax, ZOOM_ALL);
    }
  
  ax->click = ax->drag = ax->release = NULL;
  ax->c_version++;
  ax->g_version++;
  ax->status &= VIEWER_QROT;
  SetUpContext ( ax );

  if (Visible (viewer) && AllParentsVisible (viewer)) {
    tempPort = SavePort (viewer);
    Select (viewer);   
    ObjectRect ( viewer, &r );
    InvalRect (&r);
    RestorePort (tempPort);
  }

  return DiagHasErrorRec ();
}


void Nlm_ResetViewer3D ( Nlm_Viewer3D viewer )
{
  PViewer3DPtr extra;

  Nlm_DiagReset ();
  GetPanelExtra ( (PaneL)viewer, &extra);
  ResetViewerProc ( (PaneL)viewer );
}


void Nlm_GetViewerInfo3D ( Nlm_Viewer3D viewer, Picture3D PNTR pic,
                           Nlm_Camera3DPtr camera, RectPtr rectangle )
{
  register PViewer3DPtr ax;
  PViewer3DPtr   extra;
  PCamera3DUnion camUnn;

  Nlm_DiagReset ();
  GetPanelExtra ( (PaneL)viewer, &extra);
  ax = extra;
  if ( ax == NULL ) return;

  if ( pic != NULL ) *pic = (Picture3D)ax->pic;
  if (camera != NULL  &&  ax->pic != NULL)
    {
      camera->dummy[0] = CAMERA_KEY;

      MemCpy ((void*)&(camera->dummy[1] ), (void*)&(ax->a[0][0]),
	      sizeof(ax->a));
      MemCpy ((void*)&(camera->dummy[10]), (void*)&(ax->c[0]),
	      sizeof(ax->c));

      camUnn.colorHLR   = ax->colorHLR;
      camUnn.colorHLG   = ax->colorHLG;
      camUnn.colorHLB   = ax->colorHLB;
      camUnn.perspectN  = ax->perspectN;
      camUnn.brighN     = ax->brighN;
      camUnn.colorPicN  = ax->colorPicN;
      camUnn.colorFon   = ax->colorFon;
      camUnn.colorFonR  = ax->colorFonR;
      camUnn.colorFonG  = ax->colorFonG;
      camUnn.colorFonB  = ax->colorFonB;
      camUnn.adjust     = ax->adjust;
      camUnn.scale      = ax->scale;

      MemCpy ((void*)&(camera->dummy[13]), (void*)&camUnn,
	      sizeof(camUnn));
    }

  if ( rectangle != NULL ) ObjectRect ( viewer, rectangle );
}


Boolean  Nlm_RedrawViewer3D ( Nlm_Viewer3D viewer )
{
  WindoW        tempPort;
  PViewer3DPtr  extra;

  Nlm_DiagReset ();
  if (Visible (viewer) && AllParentsVisible (viewer)) {
    GetPanelExtra ( (PaneL)viewer, &extra);
    tempPort = SavePort (viewer);
    Select (viewer);
    DrawViewer3D ( (PaneL)viewer );
    RestorePort (tempPort);
  }
  return DiagHasErrorRec ();
}

void Nlm_SetViewer3DProcs ( Nlm_Viewer3D viewer, Nlm_V3DClckProc click,
                            Nlm_V3DClckProc drag, Nlm_V3DClckProc release )
{
  register PViewer3DPtr ax;
  PViewer3DPtr extra;

  Nlm_DiagReset ();
  GetPanelExtra ( (PaneL)viewer, &extra);
  ax = extra;
  if ( ax == NULL ) return;
  ax->click = click;
  ax->drag = drag;
  ax->release = release;
}

static Boolean Nlm_AddFoundPrim ( PViewer3DPtr p, Uint4 primId )
{
  Int2 i;

  if ( primId == 0 ) return TRUE;
  for (i=0; i<(Int2)p->nPrimFound; i++ ){
    if ( p->fondPrim[i] == primId ) return TRUE;
  }
  p->fondPrim[p->nPrimFound] = primId;
  p->nPrimFound++;
  if ( p->nPrimFound == 32 ) return FALSE;
  return TRUE;
}

Uint2 Nlm_FindPrim3D ( Nlm_Viewer3D viewer, PoinT pt )
{
  register PViewer3DPtr ax;
  PViewer3DPtr  extra;
  RecT          r;
  int           i,j,k;
  int           x, y;

  Nlm_DiagReset ();
  ObjectRect ( viewer, &r );
  GetPanelExtra ( (PaneL)viewer, &extra);
  ax = extra;
  if ( ax == NULL ) return 0;
  Nlm_stCon.zBuffer = (Uint2Ptr)MemNew ( (2*FIND_BLOCK_SIZE+1) * 
                                          (2*FIND_BLOCK_SIZE+1) * 2);
  Nlm_stCon.idBuffer = (Uint4Ptr)MemNew ( (2*FIND_BLOCK_SIZE+1) * 
                                          (2*FIND_BLOCK_SIZE+1) * 4);
  if ( (Nlm_stCon.zBuffer == NULL) || (Nlm_stCon.idBuffer == NULL) ) {
    Nlm_DiagPutRecord ( DA_ERROR, viewerClass, "FindPrim3D",
                         "Can not lock memory block" );
    if ( Nlm_stCon.zBuffer != NULL ) MemFree(Nlm_stCon.zBuffer);
    if ( Nlm_stCon.idBuffer != NULL ) MemFree(Nlm_stCon.idBuffer);
    return 0;
  }
  pt.x -= r.left + 2;
  pt.y -= r.top  + 4;
#ifdef NEGYSCREEN
  pt.y = (Int2)(ax->height - pt.y);
#endif
  SetUpContext ( ax );
  Nlm_stCon.widthCur = Nlm_stCon.heightCur = FIND_BLOCK_SIZE * 2 + 1;
  Nlm_UpdateGver ( ax->pic );
  Nlm_stCon.xmin = pt.x - FIND_BLOCK_SIZE;
  Nlm_stCon.xmax = pt.x + FIND_BLOCK_SIZE;
  Nlm_stCon.ymin = pt.y - FIND_BLOCK_SIZE;
  Nlm_stCon.ymax = pt.y + FIND_BLOCK_SIZE;
  ax->pic->seg.base.fTable->hittest((Nlm_VoidPtr)&(ax->pic->seg));
  ax->nPrimFound = 0;
  for ( i=0; i<=(FIND_BLOCK_SIZE-1); i++ ){
    for ( j=-i; j<=i; j++ ){
      for ( k=0; k<4; k++ ){
        switch(k){
          case 0:
            x = -j; y = -i; break;
          case 1:
            x =  j; y =  i; break;
          case 2:
            x = -i; y =  j; break;
          default:
            x =  i; y = -j;
        }
        x += FIND_BLOCK_SIZE;
        y += FIND_BLOCK_SIZE;
        Nlm_AddFoundPrim ( ax, Nlm_stCon.idBuffer[y*(FIND_BLOCK_SIZE+1)+x] );
      }
    }
  }
  MemFree (Nlm_stCon.zBuffer);
  MemFree (Nlm_stCon.idBuffer);
  return ax->nPrimFound;
}

Prim3D Nlm_GetFoundPrim3D ( Nlm_Viewer3D viewer, Uint2 primIndex )
{
  register PViewer3DPtr ax;
  PViewer3DPtr extra;

  Nlm_DiagReset ();
  GetPanelExtra ( (PaneL)viewer, &extra);
  ax = extra;
  if ( ax == NULL ) return NULL;
  if ( primIndex >= ax->nPrimFound ) return NULL;
  return (Prim3D)ax->fondPrim[primIndex];
}

void Nlm_SetHLColor3D ( Nlm_Viewer3D viewer,
                        Uint1 red, Uint1 green, Uint1 blue )
{
  register PViewer3DPtr ax;
  PViewer3DPtr extra;

  Nlm_DiagReset ();
  GetPanelExtra ( (PaneL)viewer, &extra);
  ax = extra;
  if ( ax == NULL ) return;
  ax->colorHLR = red;
  ax->colorHLG = green;
  ax->colorHLB = blue;
  ax->c_version++;
}

static void Nlm_RedrawPrim3D ( PViewer3DPtr viewer, 
                               Nlm_Base3DPtr pBase )
{
  register PViewer3DPtr   pV;
  register Nlm_PPict3DPtr pP;
  Uint2                   iwidth, iheight;
  Uint1                   color;

  pV = viewer;
  pP = pV->pic;
  if ( pP == NULL ) return;
  Nlm_stCon.image = LockPixMapImage ( pV->image );
  if ( Nlm_stCon.image == NULL ) {
    Nlm_DiagPutRecord ( DA_ERROR, viewerClass, "DrawViewer3D",
                        "Can not lock memory block" );
    return;
  }
  Nlm_stCon.zBuffer = HandLock( pV->zBuffer );
  if ( Nlm_stCon.zBuffer == NULL ) {
    Nlm_DiagPutRecord ( DA_ERROR, viewerClass, "DrawViewer3D",
                         "Can not lock memory block" );
    return;
  }
  pV->colorScale = (pV->zmax<<8)/pV->colorStep;
  GetImageSize ( pV->image, &iwidth, &iheight );
  SetUpContext ( pV );
  if ( pBase->status & HL_MASK ){
    color = pBase->color;
    pBase->color = Nlm_stCon.colorHL;
    pBase->fTable->draw ( (Nlm_VoidPtr)pBase );
    pBase->color = color;
  } else {
    pBase->fTable->draw ( (Nlm_VoidPtr)pBase );
  }
  ImageModified ( pV->image );
  pV->status |= IMAGE_READY;
  HandUnlock ( pV->zBuffer );
  UnlockPixMapImage ( pV->image );
}

void Nlm_HighlightSeg3D ( Nlm_Viewer3D viewer, Segment3D segment, 
                          Boolean highlight )
{
  register PViewer3DPtr pV;
  register Uint4        ax;
  PViewer3DPtr          extra;

  Nlm_DiagReset ();
  GetPanelExtra ( (PaneL)viewer, &extra);
  pV = extra;
  if ( pV == NULL ) return;
  ax = (Uint4)segment;
  if ( ax == 0 ) return;
  if ( highlight ) ((Nlm_Base3DPtr)ax)->status |= HL_SEGMENT;
  else ((Nlm_Base3DPtr)ax)->status &= ~HL_MASK;
  ax = (Uint4)(((Nlm_PSeg3DPtr)ax)->first);
  while ( ax ){
    if ( ((Nlm_Base3DPtr)ax)->fTable->code == 0 ){
      Nlm_HighlightSeg3D ( viewer, (Segment3D)ax, highlight );
    } else if ( highlight ) {
      ((Nlm_Base3DPtr)ax)->status |= HL_SEGMENT;
    } else {
      ((Nlm_Base3DPtr)ax)->status &= ~HL_SEGMENT;
    }
    ax = (Uint4)(((Nlm_Base3DPtr)ax)->next);
  }
  if ( segment != NULL ){
    if ( pV->pic->version == pV->p_version ) {
      Nlm_RedrawPrim3D ( pV, (Nlm_Base3DPtr)segment );
      pV->pic->version++;
      pV->p_version++;
    } else {
      pV->pic->version++;
    }
  }
}

Boolean Nlm_IsSeg3DHlighted ( Nlm_Viewer3D viewer, Segment3D segment )
{
  Nlm_DiagReset ();
  if ( (segment != NULL) && 
       (((Nlm_Base3DPtr)segment)->status & HL_SEGMENT) ) return TRUE;
  return FALSE;
}

void Nlm_HighlightPrim3D ( Nlm_Viewer3D viewer, Prim3D prim, 
                           Boolean highlight )
{
  register PViewer3DPtr pV;
  PViewer3DPtr          extra;

  Nlm_DiagReset ();
  GetPanelExtra ( (PaneL)viewer, &extra);
  pV = extra;
  if ( pV == NULL ) return;
  if ( prim != NULL ){
    if ( highlight ) {
      ((Nlm_Base3DPtr)prim)->status |= HL_PRIMITIVE;
    } else {
      ((Nlm_Base3DPtr)prim)->status &= ~HL_PRIMITIVE;
    }
    if ( pV->pic->version == pV->p_version ) {
      Nlm_RedrawPrim3D ( pV, (Nlm_Base3DPtr)prim );
      pV->pic->version++;
      pV->p_version++;
    } else {
      pV->pic->version++;
    }
  }
}

Boolean Nlm_IsPrim3DHlighted ( Nlm_Viewer3D viewer, Prim3D prim )
{
  Nlm_DiagReset ();
  if ( (prim != NULL) &&
       (((Nlm_Base3DPtr)prim)->status & HL_PRIMITIVE) ) return TRUE;
  return FALSE;
}

static void Nlm_ResetHLSeg3D ( Segment3D segment )
{
  register Uint4 ax;

  ax = (Uint4)segment;
  ((Nlm_Base3DPtr)ax)->status &= ~HL_MASK;
  ax = (Uint4)(((Nlm_PSeg3DPtr)ax)->first);
  while ( ax ){
    if ( ((Nlm_Base3DPtr)ax)->fTable->code == 0 ){
      Nlm_ResetHLSeg3D ( (Segment3D)ax );
    } else {
      ((Nlm_Base3DPtr)ax)->status &= ~HL_MASK;
    }
    ax = (Uint4)(((Nlm_Base3DPtr)ax)->next);
  }
}

Boolean Nlm_ResetHighlight3D ( Nlm_Viewer3D viewer )
{
  register PViewer3DPtr pV;
  PViewer3DPtr          extra;

  Nlm_DiagReset ();
  GetPanelExtra ( (PaneL)viewer, &extra);
  pV = extra;
  if ( (pV != NULL)&&(pV->pic != NULL) ){
    Nlm_ResetHLSeg3D ( (Segment3D)&(extra->pic->seg) );
    pV->pic->version++;
  }
  return TRUE;
}

Boolean Nlm_MapWorldToViewer3D ( Nlm_Viewer3D viewer, Int4 x, Int4 y, Int4 z,
                                 PoinT PNTR vPoint )
{
  register PViewer3DPtr ax;
  PViewer3DPtr          extra;
  Int4                  xs, ys, zs;
  Int4                  t4, zmaxPersp, zmaxPersp1;

  Nlm_DiagReset ();
  GetPanelExtra ( (PaneL)viewer, &extra);
  ax = extra;
  if ( ax == NULL ) return FALSE;
  zmaxPersp = ax->perspect * ax->zmax;
  zmaxPersp1 = zmaxPersp + ax->zmax;
  xs = x/ax->a[0][0] + y/ax->a[0][1] + z/ax->a[0][2] + ax->c[0];
  ys = x/ax->a[1][0] + y/ax->a[1][1] + z/ax->a[1][2] + ax->c[1];
  zs = x/ax->a[2][0] + y/ax->a[2][1] + z/ax->a[2][2] + ax->c[2];
  t4 = xs-(ax->width>>1);
  xs = (ax->width>>1) + (t4*zmaxPersp)/(zmaxPersp1-zs);
  t4 = ys-(ax->height>>1);
  ys = (ax->height>>1) - (t4*zmaxPersp)/(zmaxPersp1-zs);
  vPoint->x = (Int2)xs;
  vPoint->y = (Int2)ys;
  return TRUE;
}


void Nlm_SetBackground3D(Nlm_Viewer3D viewer,
                         Nlm_Uint1 red, Nlm_Uint1 green, Nlm_Uint1 blue)
{
  PViewer3DPtr  extra;
  GetPanelExtra ((PaneL)viewer, &extra);
  if (extra == NULL)
      return;

  SetColorCell((Nlm_GraphiC)extra->viewer, extra->colorFon,
               (extra->colorFonR = red),
               (extra->colorFonG = green),
               (extra->colorFonB = blue));

  extra->c_version++;
}
